@@ -4,14 +4,17 @@ import io
 import json
 import os
 import random
+import sys
 from abc import ABCMeta, abstractmethod
 from collections import deque
 from dataclasses import dataclass
-from typing import Any, Deque, Dict, List, Tuple
+from functools import wraps
+from typing import Deque, Dict, Iterable, List, Tuple
 from zipfile import ZipFile
 
 import torch
 from torch.nn import Module
+from torch.optim import Optimizer
 from torch.utils.data import DataLoader
 
 from .. import logger
@@ -27,7 +30,13 @@ from .contractor import (BenchmarkDoneEvent, CheckinEvent,
                          StartAggregatorElectionEvent, StartRoundEvent,
                          SyncStateEvent, SyncStateResponseEvent)
 
-__all__ = ['ConfigError', 'AggregationError', 'FedAvgScheduler']
+__all__ = [
+    'ConfigError',
+    'AggregationError',
+    'FedAvgScheduler',
+    'FedSGDScheduler',
+    'register_metrics'
+]
 
 
 DEV_TASK_ID = 'ceb1ddac12e54fdab888194a01137e8b'
@@ -79,6 +88,7 @@ class AggregatorSyncData:
 
 
 class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
+    """Implementation of DPFedAvg."""
 
     _INIT = 'init'
     _GETHORING = 'gethering'
@@ -111,12 +121,12 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         r"""Init.
 
         :args
-            :participants
-                ID list of all participants registered on the task.
             :min_clients
                 Minimal number of calculators for each round.
             :max_clients
                 Maximal number of calculators for each round.
+            :name
+                Default to the task ID.
             :max_rounds
                 Maximal number of training rounds.
             :merge_epochs
@@ -168,6 +178,11 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         self._save_path = os.path.join('models', self.name)
         self._round = 0
 
+        self._metrics_bucket: Dict[str, MetricsTrace] = {}
+
+        self.model = None
+        self.optimizer = None
+
     def _validate_config(self):
         if self.min_clients > self.max_clients:
             raise ConfigError('min_clients must not exceed max_clients')
@@ -178,6 +193,10 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
     def make_model(self) -> Module:
         """Return a model object which will be used for training."""
 
+    @abstractmethod
+    def make_optimizer(self) -> Optimizer:
+        """Return a optimizer object which will be used for training."""
+
     @abstractmethod
     def make_train_dataloader(self) -> DataLoader:
         """Define the training dataloader.
@@ -224,18 +243,21 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
         For example: check if train and test dataset could be loaded successfully.
         """
-        self.model = self.make_model()
         if self.model is None:
-            raise ConfigError('must specify a model to train')
+            raise ConfigError('Must specify a model to train')
         if not isinstance(self.model, Module):
-            raise ConfigError('support torch.Module only')
+            raise ConfigError('Support torch.Module only')
+        if self.optimizer is None:
+            raise ConfigError('Must specify an optimizer to train')
+        if not isinstance(self.optimizer, Optimizer):
+            raise ConfigError('Support torch.optim.Optimizer only')
 
     @abstractmethod
     def train(self):
         """Define the training steps."""
 
     @abstractmethod
-    def test(self) -> Dict[str, Any]:
+    def test(self):
         """Define the training steps.
 
         If you do not want to do testing after training, simply make it pass.
@@ -245,6 +267,11 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         """By default true if reach the max rounds configured."""
         return self._is_reach_max_rounds()
 
+    def _run(self, id: str, task_id: str, is_initiator: bool = False):
+        self._setup_context(id=id, task_id=task_id, is_initiator=is_initiator)
+        self.push_log(message='Local context is ready.')
+        self._launch_process()
+
     def _setup_context(self, id: str, task_id: str, is_initiator: bool = False):
         assert id, 'must specify a unique id for every participant'
         assert task_id, 'must specify a task_id for every participant'
@@ -259,21 +286,24 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
         self.contractor = FedAvgContractor(task_id=task_id)
         self.data_channel = GRPCDataChannel(self.contractor)
+        self.model = self.make_model()
+        self.optimizer = self.make_optimizer()
 
-        self._push_log(message='Begin to validate local context.')
+        self.push_log(message='Begin to validate local context.')
         self.validate_context()
-        self._push_log(message='Local context is ready.')
 
-    def _push_log(self, message: str):
+    def push_log(self, message: str):
         """Push a running log message to the task manager."""
-        assert message and isinstance(message, str), f'invalid log message: {message}'
+        super().push_log(message=message)
         logger.info(message)
-        self.contractor.push_log(source_id=self.id, message=message)
 
-    def _run(self):
+    def get_metrics(self, name: str) -> MetricsTrace:
+        return self._metrics_bucket.get(name)
+
+    def _launch_process(self):
         try:
             assert self.status == self._INIT, 'must begin from initial status'
-            self._push_log(f'Node {self.id} is up.')
+            self.push_log(f'Node {self.id} is up.')
 
             self.status = self._GETHORING
             logger.debug(f'{self.status=}')
@@ -299,7 +329,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 logger.debug(f'{self.status=}')
 
                 if self.is_aggregator and self.is_task_finished():
-                    self._push_log(f'Obtained the final results of task {self.task_id}')
+                    self.push_log(f'Obtained the final results of task {self.task_id}')
                     self.status = self._FINISHING
                     logger.debug(f'{self.status=}')
                     self._close_task()
@@ -315,18 +345,16 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         As a participant, checkins and gets ready for election or training.
         """
         if self.is_initiator:
-            self._push_log('Waiting for participants taking part in ...')
+            self.push_log('Waiting for participants taking part in ...')
             self._wait_for_gathering()
         else:
             is_checked_in = False
             # the aggregator may be in special state so can not response
             # correctly nor in time, then retry periodically
-            self._push_log('Checking in the task ...')
+            self.push_log('Checking in the task ...')
             while not is_checked_in:
-                nonce = self.contractor.checkin(peer_id=self.id)
-                is_checked_in = self._wait_for_checkin_response(nonce=nonce,
-                                                                timeout=self.schedule_timeout)
-        self._push_log(f'Node {self.id} have taken part in the task.')
+                is_checked_in = self._check_in_task()
+        self.push_log(f'Node {self.id} have taken part in the task.')
 
     def _sync_state(self):
         """Synchronize state before each round, so it's easier to manage the process.
@@ -334,19 +362,19 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         As an aggregator, iterates round and broadcasts, resets context of the new round.
         As a participant, synchronizes state and gives a response.
         """
-        self._push_log('Synchronizing round state ...')
+        self.push_log('Synchronizing round state ...')
         if self.is_aggregator:
             self._round += 1
             self._calculators.clear()
-            self._push_log(f'Initiate state synchronization of round {self._round}.')
+            self.push_log(f'Initiate state synchronization of round {self._round}.')
             self.contractor.sync_state(round=self._round, aggregator=self.id)
             self._wait_for_sync_response()
             if len(self._calculators) < self.min_clients:
-                self._push_log('Task failed because of too few participants.')
+                self.push_log('Task failed because of too few participants.')
                 raise AggregationError(f'too few participants: {len(self._calculators)}')
         else:
             self._wait_for_sync_state()
-        self._push_log(f'Successfully synchronized state in round {self._round}')
+        self.push_log(f'Successfully synchronized state in round {self._round}')
 
     def _elect_aggregator(self):
         """Elect a new aggregator for a round.
@@ -357,7 +385,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         As a participant, performes benchmark and reports the result. If is selected
         as the new aggregator, receives context data and takes over management.
         """
-        self._push_log('Electing a new aggregator.')
+        self.push_log('Electing a new aggregator.')
         if self.is_aggregator:
             self.contractor.start_aggregator_election(round=self._round)
             new_aggregator = self._wait_for_benchmark()
@@ -365,11 +393,11 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             # it's possible the all benchmark failed
             self.is_aggregator = new_aggregator == self.id
         else:
-            self._push_log('Waiting for aggregator election begin ...')
+            self.push_log('Waiting for aggregator election begin ...')
             self._wait_for_electing()
-            self._push_log('Waiting for aggregator election result ...')
+            self.push_log('Waiting for aggregator election result ...')
             self._wait_for_election_result()
-        self._push_log('Aggregator election complete.')
+        self.push_log('Aggregator election complete.')
 
     def _run_a_round(self):
         """Perform a round of FedAvg calculation.
@@ -392,13 +420,13 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         uploads the final parameters and tells L1 task manager the task is complete.
         As a participant, do nothing.
         """
-        self._push_log(f'Closing task {self.task_id} ...')
+        self.push_log(f'Closing task {self.task_id} ...')
         if self.is_aggregator:
             self.status = self._FINISHING
             logger.debug(f'{self.status=}')
             self.contractor.finish_task()
             self._upload_task_output()
-        self._push_log(f'Task {self.task_id} closed. Byebye!')
+        self.push_log(f'Task {self.task_id} closed. Byebye!')
 
     def _wait_for_gathering(self):
         """Wait for participants gethering."""
@@ -407,13 +435,19 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             if isinstance(_event, CheckinEvent):
                 self._handle_check_in(_event)
                 if len(self._participants) >= self.min_clients:
-                    return
+                    break
+        # it takes seconds to send a response and to consume it in the remote side,
+        # meanwhile, more check in events may arrive soonly. So keep looking for a while.
+        countdown = 10
+        for _event in self.contractor.contract_events(timeout=countdown):
+            if isinstance(_event, CheckinEvent):
+                self._handle_check_in(_event)
 
     def _handle_check_in(self, _event: CheckinEvent):
         if _event.peer_id not in self._participants:
             self._participants.append(_event.peer_id)
-            self._push_log(f'Welcome a new participant ID: {_event.peer_id}.')
-            self._push_log(f'There are {len(self._participants)} participants now.')
+            self.push_log(f'Welcome a new participant ID: {_event.peer_id}.')
+            self.push_log(f'There are {len(self._participants)} participants now.')
         self.contractor.respond_checkin(round=self._round,
                                         aggregator=self.id,
                                         nonce=_event.nonce,
@@ -421,20 +455,26 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
     def _wait_for_sync_response(self):
         """Wait for participants' synchronizing state response."""
-        self._push_log('Waiting for synchronization responses ...')
+        self.push_log('Waiting for synchronization responses ...')
         for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
             if isinstance(_event, SyncStateResponseEvent):
                 if _event.round != self._round:
                     continue
                 self._calculators.append(_event.peer_id)
-                self._push_log(f'Successfully synchronized state with ID: {_event.peer_id}.')
-                self._push_log(f'Successfully synchronized with {len(self._calculators)} participants.')
+                self.push_log(f'Successfully synchronized state with ID: {_event.peer_id}.')
+                self.push_log(f'Successfully synchronized with {len(self._calculators)} participants.')
                 if len(self._calculators) == len(self._participants):
                     return
 
             elif isinstance(_event, CheckinEvent):
                 self._handle_check_in(_event)
 
+    def _check_in_task(self) -> bool:
+        """Try to check in the task."""
+        nonce = self.contractor.checkin(peer_id=self.id)
+        return self._wait_for_checkin_response(nonce=nonce,
+                                               timeout=self.schedule_timeout)
+
     def _wait_for_checkin_response(self, nonce: str, timeout: int = 0) -> bool:
         """Wait for checkin response.
 
@@ -456,14 +496,14 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
         Return True if synchronising successfully otherwise False.
         """
-        self._push_log('Waiting for synchronizing state with the aggregator ...')
+        self.push_log('Waiting for synchronizing state with the aggregator ...')
         for _event in self.contractor.contract_events(timeout=timeout):
             if isinstance(_event, SyncStateEvent):
                 self._round = _event.round
                 self.contractor.respond_sync_state(round=self._round,
                                                    peer_id=self.id,
                                                    aggregator=_event.aggregator)
-                self._push_log('Successfully synchronized state with the aggregator.')
+                self.push_log('Successfully synchronized state with the aggregator.')
                 return
             elif isinstance(_event, FinishTaskEvent):
                 raise _TaskComplete()
@@ -474,7 +514,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         :return
             the ID of the newly selected aggregator
         """
-        self._push_log('Waiting for benchmark results ...')
+        self.push_log('Waiting for benchmark results ...')
         new_aggregator = self.id
         self._calculators.clear()
         for _event in self.contractor.contract_events(timeout=SimplePerfBench.TIMEOUT):
@@ -484,13 +524,13 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
                 if _event.peer_id not in self._calculators:
                     self._calculators.append(_event.peer_id)
-                    self._push_log(f'Received benchmark result from ID: {_event.peer_id}.')
+                    self.push_log(f'Received benchmark result from ID: {_event.peer_id}.')
 
                 # the first one arrived
                 if new_aggregator == self.id:
                     new_aggregator = _event.peer_id
-                    self._push_log(f'ID: {new_aggregator} is elected as the new aggregator.')
-                    self._push_log('Begin to synchronize context with the new aggregatro.')
+                    self.push_log(f'ID: {new_aggregator} is elected as the new aggregator.')
+                    self.push_log('Begin to synchronize context with the new aggregatro.')
 
                     buffer = io.BytesIO()
                     torch.save(self.state_dict(), buffer)
@@ -499,7 +539,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                     self.data_channel.send_stream(source=self.id,
                                                   target=_event.peer_id,
                                                   data_stream=sync_data.to_bytes())
-                    self._push_log('Successfully synchronized context with the new aggregatro.')
+                    self.push_log('Successfully synchronized context with the new aggregatro.')
 
                 if len(self._calculators) == len(self._participants):  # all finished
                     break
@@ -528,8 +568,8 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         for _event in self.contractor.contract_events():
             if isinstance(_event, ApplySendingDataEvent):
                 # is selected as the next aggregator and synchronise aggregator data
-                self._push_log('Current node is selected as the new aggregator.')
-                self._push_log('Begin to synchronize context with the original aggregator.')
+                self.push_log('Current node is selected as the new aggregator.')
+                self.push_log('Begin to synchronize context with the original aggregator.')
                 data_stream = self.data_channel.receive_stream(session_id=_event.session_id,
                                                                source=_event.source,
                                                                sender_key=_event.public_key)
@@ -541,7 +581,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 if self.id in self._participants:
                     self._participants.remove(self.id)
                 self.is_aggregator = True
-                self._push_log('Successfully synchronized context with the original aggregator.')
+                self.push_log('Successfully synchronized context with the original aggregator.')
 
             elif isinstance(_event, CloseAggregatorElectionEvent):
                 if _event.round != self._round:
@@ -553,20 +593,39 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 raise _TaskComplete()
 
     def _run_as_aggregator(self):
-        self._push_log(f'Begin the training of round {self._round}.')
+        self._start_round()
+        self._distribute_model()
+
+        self._process_aggregation()
+
+        self._persistent_model()
+        self._check_and_run_test()
+        self._close_round()
+
+    def _close_round(self):
+        """Close current round when finished."""
+        self.status = self._CLOSING_ROUND
+        logger.debug(f'{self.status=}')
+        self.contractor.close_round(round=self._round)
+        self.push_log(f'The training of Round {self._round} complete.')
+
+    def _start_round(self):
+        """Prepare and start calculation of a round."""
+        self.push_log(f'Begin the training of round {self._round}.')
         self.status = self._SELECTING
         logger.debug(f'{self.status=}')
-        self._push_log('Select calculators of current round.')
+        self.push_log('Select calculators of current round.')
         self._select_calculators()
-        self._push_log(f'Calculators of round {self._round} are: {self._calculators}.')
+        self.push_log(f'Calculators of round {self._round} are: {self._calculators}.')
         self.contractor.start_round(round=self._round,
                                     calculators=self._calculators,
                                     aggregator=self.id)
-        self._push_log(f'Calculation of round {self._round} is started.')
+        self.push_log(f'Calculation of round {self._round} is started.')
 
+    def _distribute_model(self):
         buffer = io.BytesIO()
         torch.save(self.state_dict(), buffer)
-        self._push_log('Distributing parameters ...')
+        self.push_log('Distributing parameters ...')
         results = {_target: False for _target in self._calculators}
         connection_timeout, timeout = self.data_channel_timeout
         for _target in self._calculators:
@@ -577,28 +636,30 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                                               connection_timeout=connection_timeout,
                                               timeout=timeout)
                 results[_target] = True
-                self._push_log(f'Successfully distributed parameters to ID: {_target}')
+                self.push_log(f'Successfully distributed parameters to ID: {_target}')
             except SendingError:
-                self._push_log(f'Failed to distribute parameters to ID: {_target}')
+                self.push_log(f'Failed to distribute parameters to ID: {_target}')
                 logger.exception(f'failed to send parameters to {_target}')
         if sum(results.values()) < self.min_clients:
-            self._push_log('Task failed because of too few calculators getting ready')
+            self.push_log('Task failed because of too few calculators getting ready')
             raise AggregationError(f'Too few calculators getting ready: {results}.')
-        self._push_log(f'Distributed parameters to {sum(results.values())} calculators.')
+        self.push_log(f'Distributed parameters to {sum(results.values())} calculators.')
 
+    def _process_aggregation(self):
+        """Process aggregation depending on specific algorithm."""
         self.status = self._WAIT_FOR_AGGR
         logger.debug(f'{self.status=}')
         self.contractor.notify_ready_for_aggregation(round=self._round)
-        self._push_log('Now waiting for executing calculation ...')
+        self.push_log('Now waiting for executing calculation ...')
         accum_result, result_count = self._wait_for_calculation()
         if result_count < self.min_clients:
-            self._push_log('Task failed because of too few calculation results gathered.')
+            self.push_log('Task failed because of too few calculation results gathered.')
             raise AggregationError(f'Too few results gathered: {result_count} copies.')
-        self._push_log(f'Received {result_count} copies of calculation results.')
+        self.push_log(f'Received {result_count} copies of calculation results.')
 
         self.status = self._AGGREGATING
         logger.debug(f'{self.status=}')
-        self._push_log('Begin to aggregate and update parameters.')
+        self.push_log('Begin to aggregate and update parameters.')
         for _key in accum_result.keys():
             if accum_result[_key].dtype in (
                 torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64
@@ -608,29 +669,23 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             else:
                 accum_result[_key].div_(result_count)
         self.load_state_dict(accum_result)
-        self._push_log('Obtained a new version of parameters.')
+        self.push_log('Obtained a new version of parameters.')
 
+    def _persistent_model(self):
         self.status = self._PERSISTING
         logger.debug(f'{self.status=}')
         self._save_model()
 
+    def _check_and_run_test(self):
+        """Run test if match configured conditions."""
         if (
             self._round == 1
             or (self.log_rounds > 0 and self._round % self.log_rounds == 0)
             or self._round == self.max_rounds
         ):
-            self._push_log('Begin to make a model test.')
+            self.push_log('Begin to make a model test.')
             test_result = self.test()
-            if self._round == 1:
-                self._default_metrics = MetricsTrace(headers=test_result.keys())
-                self.register_metrics(name='test_results', metrics=self._default_metrics)
-            self._default_metrics.append_metrics_item(test_result)
-            self._push_log(f'The latest test results are: {test_result}')
-
-        self.status = self._CLOSING_ROUND
-        logger.debug(f'{self.status=}')
-        self.contractor.close_round(round=self._round)
-        self._push_log(f'The training of Round {self._round} complete.')
+            self.push_log(f'The latest test results are: {test_result}')
 
     def _wait_for_calculation(self) -> Tuple[Dict[str, torch.Tensor], int]:
         """Wait for every calculator finish its task or timeout."""
@@ -640,7 +695,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             if isinstance(_param, torch.Tensor):
                 _param.zero_()
 
-        self._push_log('Waiting for calculation results ...')
+        self.push_log('Waiting for calculation results ...')
         for _event in self.contractor.contract_events(timeout=self.calculation_timeout):
             if isinstance(_event, ApplySendingDataEvent):
                 if _event.target != self.id:
@@ -653,16 +708,13 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 for _key in accum_result.keys():
                     accum_result[_key].add_(_new_state_dict[_key])
                 result_count += 1
-                self._push_log(f'Received calculation results from ID: {_event.source}')
+                self.push_log(f'Received calculation results from ID: {_event.source}')
                 if result_count >= len(self._calculators):
                     return accum_result, result_count
 
             elif isinstance(_event, CheckinEvent):
                 self._handle_check_in(_event)
 
-            elif isinstance(_event, FinishTaskEvent):
-                raise _TaskComplete()
-
         return accum_result, result_count
 
     def _is_reach_max_rounds(self) -> bool:
@@ -671,9 +723,8 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
     def _select_calculators(self):
         """Select calculators of a round."""
-        assert (
-            len(self._participants) >= self.min_clients
-        ), f'too few participants: {len(self._participants)}'
+        if len(self._participants) < self.min_clients:
+            raise AggregationError(f'too few participants: {len(self._participants)}')
 
         if len(self._calculators) < self.min_clients:
             candidates = [_peer for _peer in self._participants
@@ -690,7 +741,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         os.makedirs(self._save_path, exist_ok=True)
         with open(f'{os.path.join(self._save_path, "model.pt")}', 'wb') as f:
             torch.save(self.state_dict(), f)
-        self._push_log('Saved latest parameters locally.')
+        self.push_log('Saved latest parameters locally.')
 
     def _upload_task_output(self) -> Tuple[str, str]:
         """Upload final output of the task to task manager after training.
@@ -698,7 +749,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         :return
             Local paths of the report file and model file.
         """
-        self._push_log('Uploading task achievement and closing task ...')
+        self.push_log('Uploading task achievement and closing task ...')
         metrics_files = []
         for _name, _metrics in self._metrics_bucket.items():
             _file = f'{os.path.join(self._save_path, _name)}.csv'
@@ -732,43 +783,46 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         self._save_model()
         self.status = self._CALCULATING
         logger.debug(f'{self.status=}')
-        self._push_log('Begin to run calculation ...')
-        for _ in range(self.merge_epochs):
-            self.train()
-        self._push_log('Local calculation complete.')
+        self.push_log('Begin to run calculation ...')
+        self._execute_training()
+        self.push_log('Local calculation complete.')
 
         self._wait_for_uploading_model()
         buffer = io.BytesIO()
         torch.save(self.state_dict(), buffer)
         connection_timeout, timeout = self.data_channel_timeout
-        self._push_log('Pushing local update to the aggregator ...')
+        self.push_log('Pushing local update to the aggregator ...')
         self.data_channel.send_stream(source=self.id,
                                       target=self._aggregator,
                                       data_stream=buffer.getvalue(),
                                       connection_timeout=connection_timeout,
                                       timeout=timeout)
-        self._push_log('Successfully pushed local update to the aggregator.')
+        self.push_log('Successfully pushed local update to the aggregator.')
         self.status = self._CLOSING_ROUND
         logger.debug(f'{self.status=}')
         self._wait_for_closing_round()
-        self._push_log(f'ID: {self.id} finished training task of round {self._round}.')
+        self.push_log(f'ID: {self.id} finished training task of round {self._round}.')
+
+    def _execute_training(self):
+        for _ in range(self.merge_epochs):
+            self.train()
 
     def _wait_for_starting_round(self):
         """Wait for starting a new round of training."""
-        self._push_log(f'Waiting for training of round {self._round} begin ...')
+        self.push_log(f'Waiting for training of round {self._round} begin ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, StartRoundEvent):
                 if _event.round != self._round or self.id not in _event.calculators:
                     continue
                 self._aggregator = _event.aggregator
-                self._push_log(f'Training of round {self._round} begins.')
+                self.push_log(f'Training of round {self._round} begins.')
                 return
             elif isinstance(_event, FinishTaskEvent):
                 raise _TaskComplete()
 
     def _wait_for_updating_model(self):
         """Wait for receiving latest parameters from aggregator."""
-        self._push_log('Waiting for receiving latest parameters from the aggregator ...')
+        self.push_log('Waiting for receiving latest parameters from the aggregator ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, ApplySendingDataEvent):
                 if _event.target != self.id:
@@ -779,14 +833,14 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 buffer = io.BytesIO(parameters)
                 new_state_dict = torch.load(buffer)
                 self.load_state_dict(new_state_dict)
-                self._push_log('Successfully received latest parameters.')
+                self.push_log('Successfully received latest parameters.')
                 return
             elif isinstance(_event, FinishTaskEvent):
                 raise _TaskComplete()
 
     def _wait_for_uploading_model(self):
         """Wait for uploading trained parameters to aggregator."""
-        self._push_log('Waiting for aggregation begin ...')
+        self.push_log('Waiting for aggregation begin ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, ReadyForAggregationEvent):
                 if _event.round != self._round:
@@ -797,7 +851,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
     def _wait_for_closing_round(self):
         """Wait for closing current round of training."""
-        self._push_log(f'Waiting for closing signal of training round {self._round}.')
+        self.push_log(f'Waiting for closing signal of training round {self._round} ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, CloseRoundEvent):
                 if _event.round != self._round:
@@ -805,3 +859,112 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 return
             elif isinstance(_event, FinishTaskEvent):
                 raise _TaskComplete()
+
+
+def register_metrics(name: str, keys: Iterable[str]):
+    """Register a metrics trace.
+
+    The trace fo the metrics during training will be included in the task's
+    running result and can be downloaded. However, the trace content should
+    be appended manually.
+
+    If the specified name has already registered before, nothing changes.
+
+    :args
+        :name
+            The name of the metrics trace.
+        :keys
+            The names of the metrics items.
+    """
+    def decorate(func):
+        if not name or not isinstance(name, str):
+            raise ValueError(f'invalid name of the metrics: {name}')
+        if not keys or not isinstance(keys, Iterable):
+            raise ValueError(f'invalid keys of the metrics: {keys}')
+
+        @wraps(func)
+        def wrapper(_self, *args, **kwargs):
+            if not isinstance(_self, FedAvgScheduler) or func.__name__ != 'test':
+                raise RuntimeError(
+                    'Only applicable on "test" function of a "FedAvgScheduler" object.'
+                )
+            if name not in _self._metrics_bucket:
+                _self._metrics_bucket[name] = MetricsTrace(headers=keys)
+            return func(_self, *args, **kwargs)
+
+        return wrapper
+    return decorate
+
+
+class FedSGDScheduler(FedAvgScheduler):
+    """Implementation of FedSGD."""
+
+    def __init__(self,
+                 min_clients: int,
+                 name: str = None,
+                 max_rounds: int = 0,
+                 calculation_timeout: int = 300,
+                 perf_bench_timeout: int = 30,
+                 schedule_timeout: int = 30,
+                 data_channel_timeout: Tuple[int, int] = (30, 60),
+                 log_rounds: int = 0,
+                 is_centralized: bool = True):
+        r"""Init.
+
+        :args
+            :min_clients
+                Minimal number of calculators for each round.
+            :name
+                Default to the task ID.
+            :max_rounds
+                Maximal number of training rounds.
+            :calculation_timeout
+                Seconds to timeout for calculation in a round. Takeing off timeout
+                by setting its value to 0.
+            :perf_bench_timeout
+                Seconds to timeout for performing performance benchmark. Takeing off timeout
+                by setting its value to 0.
+            :schedule_timeout
+                Seconds to timeout for process scheduling. Takeing off timeout
+                by setting its value to 0.
+            :data_channel_timeout
+                NOTE: This is a temporary configuration. May be removed in next version.
+                Do not design your process depending on it.\n
+                A pair of timeout configuration. The first is seconds to timeout for
+                connecting, and the second is seconds to timeout for transmitting data.
+                Takeing off either one by setting its value to 0.
+            :log_rounds
+                The number of rounds to run testing and log the result. Skip it
+                by setting its value to 0.
+            :is_centralized
+                If specify centralized, the aggregator will always be the initiator of the
+                task, otherwize a new aggregator is elected for each round during training.
+        """
+        super().__init__(min_clients=min_clients,
+                         max_clients=sys.maxsize,
+                         name=name,
+                         max_rounds=max_rounds,
+                         merge_epochs=1,
+                         calculation_timeout=calculation_timeout,
+                         perf_bench_timeout=perf_bench_timeout,
+                         schedule_timeout=schedule_timeout,
+                         data_channel_timeout=data_channel_timeout,
+                         log_rounds=log_rounds,
+                         is_centralized=is_centralized)
+
+    def _setup_context(self, id: str, task_id: str, is_initiator: bool = False):
+        super()._setup_context(id=id, task_id=task_id, is_initiator=is_initiator)
+        # Since make_train_dataloader is implemented by the user, and DataLoader base
+        # implementation does not allow to modify batch_size after initialization,
+        # there is no way to mandatorily set batch size to the correct value.
+        # So a post check is used.
+        train_loader = self.make_train_dataloader()
+        self.push_log(f'train batch size = {train_loader.batch_size}')
+        self.push_log(f'train examples = {len(train_loader.dataset)}')
+        if train_loader.batch_size != len(train_loader.dataset):
+            raise ConfigError('batch size must be the total number of samples.')
+
+    def _select_calculators(self):
+        """Select calculators of a round."""
+        if len(self._calculators) < self.min_clients:
+            raise AggregationError(f'too few calculators: {len(self._calculators)}')
