@@ -1,4 +1,4 @@
-"""Secure FedAvg scheduler.
+"""Secure Aggregation FedAvg scheduler.
 
 Reference: https://eprint.iacr.org/2017/281.pdf
 """
@@ -17,8 +17,6 @@ from numpy.random import Generator
 
 from .. import logger
 from ..contractor import ApplySendingDataEvent
-from ..data_channel import SendingError
-from ..metrics_trace import MetricsTrace
 from ..secure import aes
 from ..secure.ecdhe import new_key_pair, new_shared_key
 from ..secure.shamir import SecretShare
@@ -32,6 +30,8 @@ from .secure_contractor import (AdvertiseKeysEvent, AdvertiseKeysResponseEvent,
                                 StartUnmaskingEvent, UploadSeedSharesEvent,
                                 UploadSKSharesEvent)
 
+__all__ = ['SecureFedAvgScheduler']
+
 
 @dataclass
 class _User1Info:
@@ -262,6 +262,9 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
         self._t = t
 
+        self._validate_secure_config()
+
+    def _validate_secure_config(self):
         # if self.min_clients < 4:
         if self.min_clients < 3:  # TODO it should be 4
             raise ConfigError('secure mode works only with at least 4 participants')
@@ -282,74 +285,41 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         else:
             return len(self._users_2)
 
-    def _run_as_aggregator(self):
-        self._push_log(f'Begin the training of round {self._round}.')
-        self.status = self._SELECTING
-        logger.debug(f'{self.status=}')
-        self._push_log('Select calculators of current round.')
-        self._select_calculators()
-        self._push_log(f'Calculators of round {self._round} are: {self._calculators}.')
-        self.contractor.start_round(round=self._round,
-                                    calculators=self._calculators,
-                                    aggregator=self.id)
-        self._push_log(f'Calculation of round {self._round} is started.')
-
-        buffer = io.BytesIO()
-        torch.save(self.state_dict(), buffer)
-        self._push_log('Distributing parameters ...')
-        results = {_target: False for _target in self._calculators}
-        connection_timeout, timeout = self.data_channel_timeout
-        for _target in self._calculators:
-            try:
-                self.data_channel.send_stream(source=self.id,
-                                              target=_target,
-                                              data_stream=buffer.getvalue(),
-                                              connection_timeout=connection_timeout,
-                                              timeout=timeout)
-                results[_target] = True
-                self._push_log(f'Successfully distributed parameters to ID: {_target}')
-            except SendingError:
-                self._push_log(f'Failed to distribute parameters to ID: {_target}')
-                logger.exception(f'failed to send parameters to {_target}')
-        if sum(results.values()) < self.min_clients:
-            self._push_log('Task failed because of too few calculators getting ready')
-            raise AggregationError(f'Too few calculators getting ready: {results}.')
-        self._push_log(f'Distributed parameters to {sum(results.values())} calculators.')
-
+    def _process_aggregation(self):
         self.status = self._ADVERTISE_KEYS
         logger.debug(f'{self.status=}')
-        self._push_log('Initiate advertising keys.')
+        self.push_log('Initiate advertising keys.')
         self.contractor.advertise_keys(round=self._round, calculators=self._calculators)
         self._wait_for_responses_of_advertising_keys()
-        self._push_log(f'Gathered user_1 list: {[_user.peer_id for _user in self._users_1]}')
+        self.push_log(f'Gathered user_1 list: {[_user.peer_id for _user in self._users_1]}')
         list_data = [_user.encode() for _user in self._users_1]
         self.contractor.distribute_user_1_list(round=self._round,
                                                list_data=list_data,
                                                targets=[_user.peer_id for _user in self._users_1])
-        self._push_log('Distributed user_1 list.')
+        self.push_log('Distributed user_1 list.')
 
         self.status = self._SHARE_KEYS
         logger.debug(f'{self.status=}')
         self._wait_for_sharing_keys()
-        self._push_log(f'Gathered user_2 list: {[_user.source_id for _user in self._users_2]}')
+        self.push_log(f'Gathered user_2 list: {[_user.source_id for _user in self._users_2]}')
         list_data = [_user.encode() for _user in self._users_2]
         self.contractor.distribute_user_2_list(
             round=self._round,
             list_data=list_data,
             targets=list(set(_user.source_id for _user in self._users_2))
         )
-        self._push_log('Distributed user_2 list.')
+        self.push_log('Distributed user_2 list.')
 
         self.status = self._MASKED_INPUT_COLLECTION
         logger.debug(f'{self.status=}')
         self.contractor.notify_ready_for_aggregation(round=self._round)
-        self._push_log('Now waiting for executing calculation ...')
+        self.push_log('Now waiting for executing calculation ...')
         accum_result, result_count = self._wait_for_calculation()
-        self._push_log(f'Gathered user_3 list: {self._users_3}')
+        self.push_log(f'Gathered user_3 list: {self._users_3}')
 
         self.status = self._UNMASKING
         logger.debug(f'{self.status=}')
-        self._push_log('Trying to unmask parameters ...')
+        self.push_log('Trying to unmask parameters ...')
         self.contractor.start_unmasking(round=self._round,
                                         list_data=self._users_3,
                                         targets=self._users_3)
@@ -359,31 +329,9 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         logger.debug(f'{self.status=}')
         self._aggregate_results(accum_result=accum_result, result_count=result_count)
 
-        self.status = self._PERSISTING
-        logger.debug(f'{self.status=}')
-        self._save_model()
-
-        if (
-            self._round == 1
-            or (self.log_rounds > 0 and self._round % self.log_rounds == 0)
-            or self._round == self.max_rounds
-        ):
-            self._push_log('Begin to make a model test.')
-            test_result = self.test()
-            if self._round == 1:
-                self._default_metrics = MetricsTrace(headers=test_result.keys())
-                self.register_metrics(name='test_results', metrics=self._default_metrics)
-            self._default_metrics.append_metrics_item(test_result)
-            self._push_log(f'The latest test results are: {test_result}')
-
-        self.status = self._CLOSING_ROUND
-        logger.debug(f'{self.status=}')
-        self.contractor.close_round(round=self._round)
-        self._push_log(f'The training of Round {self._round} complete.')
-
     def _wait_for_responses_of_advertising_keys(self):
         """Wait for responses of advertising keys as an aggregator."""
-        self._push_log('Waiting for receiving advertised keys ...')
+        self.push_log('Waiting for receiving advertised keys ...')
         self._users_1.clear()
         advertised_ids = []
         for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
@@ -396,7 +344,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                                 c_pk=_event.c_pk,
                                                 s_pk=_event.s_pk))
                 advertised_ids.append(_event.peer_id)
-                self._push_log(f'Received advertised keys from ID: {_event.peer_id}.')
+                self.push_log(f'Received advertised keys from ID: {_event.peer_id}.')
                 if len(self._users_1) == len(self._calculators):
                     break
 
@@ -404,12 +352,12 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                 self._handle_check_in(_event)
 
         if len(self._users_1) < self._t:
-            self._push_log('Task failed because of too few users_1.')
+            self.push_log('Task failed because of too few users_1.')
             raise AggregationError(f'too few users_1: {len(self._users_1)}')
 
     def _wait_for_sharing_keys(self):
         """Wait for shared keys uploaded by calculators."""
-        self._push_log('Waiting for receiving shared keys from users_1 ...')
+        self.push_log('Waiting for receiving shared keys from users_1 ...')
         self._users_2.clear()
         shared_ids = []
         for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
@@ -420,7 +368,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                     continue
                 self._users_2.extend(_User2Info.decode(_item) for _item in _event.list_data)
                 shared_ids.append(_event.peer_id)
-                self._push_log(f'Received shared keys from ID: {_event.peer_id}')
+                self.push_log(f'Received shared keys from ID: {_event.peer_id}')
                 if len(shared_ids) == len(self._users_1):
                     break
 
@@ -428,7 +376,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                 self._handle_check_in(_event)
 
         if len(shared_ids) < self._t:
-            self._push_log('Task failed because of too few users_1.')
+            self.push_log('Task failed because of too few users_1.')
             raise AggregationError(f'too few users_2: {len(shared_ids)}')
 
     def _wait_for_calculation(self) -> Tuple[Dict[str, torch.Tensor], int]:
@@ -442,7 +390,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
             else:
                 logger.warn(f'not Tensor value: {_key=}; {_param=}; {type(_param)=}')
 
-        self._push_log('Waiting for calculation results ...')
+        self.push_log('Waiting for calculation results ...')
         for _event in self.contractor.contract_events(timeout=self.calculation_timeout):
             if isinstance(_event, ApplySendingDataEvent):
                 if _event.target != self.id:
@@ -455,7 +403,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                 for _key in accum_result.keys():
                     accum_result[_key].add_(_new_state_dict[_key])
                 self._users_3.append(_event.source)
-                self._push_log(f'Received calculation results from ID: {_event.source}')
+                self.push_log(f'Received calculation results from ID: {_event.source}')
 
                 if len(self._users_3) >= user_2_num:
                     return accum_result, len(self._users_3)
@@ -464,7 +412,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                 self._handle_check_in(_event)
 
         if len(self._users_3) < self._t:
-            self._push_log('Task failed because of too few users_1.')
+            self.push_log('Task failed because of too few users_1.')
             raise AggregationError(f'too few user_3: {len(self._users_3)}')
         return accum_result, len(self._users_3)
 
@@ -474,7 +422,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self._sk_shares.clear()
         user_2_num = self._get_user_2_num()
 
-        self._push_log('Waiting for collecting secret shares ...')
+        self.push_log('Waiting for collecting secret shares ...')
         for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
             if isinstance(_event, UploadSeedSharesEvent):
                 if _event.round != self._round:
@@ -485,7 +433,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                     if _info.peer_id not in self._seed_shares:
                         self._seed_shares[_info.peer_id] = []
                     self._seed_shares[_info.peer_id].append(_info.seed_share)
-                self._push_log('Received a copy of seed shares.')
+                self.push_log('Received a copy of seed shares.')
 
                 if (
                     len(self._seed_shares) + len(self._sk_shares) == user_2_num
@@ -505,7 +453,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                     if _info.peer_id not in self._sk_shares:
                         self._sk_shares[_info.peer_id] = []
                     self._sk_shares[_info.peer_id].append(_info.sk_share)
-                self._push_log('Received a copy of SK shares.')
+                self.push_log('Received a copy of SK shares.')
 
                 if (
                     len(self._seed_shares) + len(self._sk_shares) == user_2_num
@@ -521,11 +469,11 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
     def _aggregate_results(self, accum_result: Dict[str, torch.Tensor], result_count: int):
         """Unmask and aggregate results."""
-        self._push_log('Begin to aggregate and update parameters.')
+        self.push_log('Begin to aggregate and update parameters.')
         seed_map = self._decrypt_seeds()
-        self._push_log(f'Decrypted {len(seed_map)} seeds.')
+        self.push_log(f'Decrypted {len(seed_map)} seeds.')
         sk_map = self._decrypt_sks()
-        self._push_log(f'Decrypted {len(sk_map)} sks.')
+        self.push_log(f'Decrypted {len(sk_map)} sks.')
         accum_result = self._unmask_state_dict(state_dict=accum_result,
                                                seed_map=seed_map,
                                                sk_map=sk_map)
@@ -533,7 +481,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         for _key in accum_result.keys():
             accum_result[_key].div_(result_count)
         self.load_state_dict(accum_result)
-        self._push_log('Obtained a new version of parameters.')
+        self.push_log('Obtained a new version of parameters.')
 
     def _decrypt_seeds(self) -> Dict[str, bytes]:
         assert (
@@ -587,7 +535,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
             )
         ), f'invalid state_dict data: {state_dict}'
 
-        self._push_log('Unmasking parameters ...')
+        self.push_log('Unmasking parameters ...')
         # in order to make every calculator's PRG generates masks in a same order
         sorted_keys = sorted(state_dict)
         seed_PRG_map: Dict[str, Generator] = {
@@ -628,7 +576,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                     else:
                         state_dict[_key] -= _user_mask
 
-        self._push_log('Successfully Unmasked parameters.')
+        self.push_log('Successfully Unmasked parameters.')
         return state_dict
 
     def _run_as_data_owner(self):
@@ -651,16 +599,16 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                    peer_id=self.id,
                                    list_data=shared_data,
                                    aggregator=self._aggregator)
-        self._push_log('Uploaded secret shares.')
+        self.push_log('Uploaded secret shares.')
         self._wait_for_user_2_list()
 
         self.status = self._CALCULATING
         logger.debug(f'{self.status=}')
-        self._push_log('Begin to run calculation ...')
+        self.push_log('Begin to run calculation ...')
         for _ in range(self.merge_epochs):
             self.train()
         state_dict = self.state_dict()
-        self._push_log('Local calculation complete.')
+        self.push_log('Local calculation complete.')
 
         self.status = self._MASKED_INPUT_COLLECTION
         logger.debug(f'{self.status=}')
@@ -670,13 +618,13 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         buffer = io.BytesIO()
         torch.save(masked_result, buffer)
         connection_timeout, timeout = self.data_channel_timeout
-        self._push_log('Pushing local update to the aggregator ...')
+        self.push_log('Pushing local update to the aggregator ...')
         self.data_channel.send_stream(source=self.id,
                                       target=self._aggregator,
                                       data_stream=buffer.getvalue(),
                                       connection_timeout=connection_timeout,
                                       timeout=timeout)
-        self._push_log('Successfully pushed local update to the aggregator.')
+        self.push_log('Successfully pushed local update to the aggregator.')
 
         self.status = self._UNMASKING
         logger.debug(f'{self.status=}')
@@ -685,11 +633,11 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self.status = self._CLOSING_ROUND
         logger.debug(f'{self.status=}')
         self._wait_for_closing_round()
-        self._push_log(f'ID: {self.id} finished training task of round {self._round}.')
+        self.push_log(f'ID: {self.id} finished training task of round {self._round}.')
 
     def _wait_for_advertising_keys(self):
         """Wait for the start signal and then advertise keys as a calculator."""
-        self._push_log('Waiting for the signal to advertise keys ...')
+        self.push_log('Waiting for the signal to advertise keys ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, AdvertiseKeysEvent):
                 if _event.round != self._round or self.id not in _event.calculators:
@@ -701,7 +649,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                                        peer_id=self.id,
                                                        c_pk=self._c_pk,
                                                        s_pk=self._s_pk)
-                self._push_log(f'ID: {self.id} has advertised keys.')
+                self.push_log(f'ID: {self.id} has advertised keys.')
                 return
             elif isinstance(_event, FinishTaskEvent):
                 raise _TaskComplete()
@@ -709,7 +657,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
     def _wait_for_user_1_list(self):
         """Wait for receiving the list of user_1."""
         self._users_1.clear()
-        self._push_log('Waiting for receiving users_1 list ...')
+        self.push_log('Waiting for receiving users_1 list ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, DistributeUser1ListEvent):
                 if _event.round != self._round:
@@ -718,16 +666,16 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                  for _user in _event.list_data]
                 self._users_1 = [_user for _user in self._users_1
                                  if _user.peer_id != self.id]
-                self._push_log(f'Received {len(self._users_1)} users_1.')
+                self.push_log(f'Received {len(self._users_1)} users_1.')
 
                 if len(self._users_1) < self._t - 1:
-                    self._push_log('Task failed because of too few users_1.')
+                    self.push_log('Task failed because of too few users_1.')
                     raise AggregationError(f'too few users_1: {len(self._users_1)}')
                 if (
                     len(set(_user.c_pk for _user in self._users_1)) < len(self._users_1)
                     or len(set(_user.s_pk for _user in self._users_1)) < len(self._users_1)
                 ):
-                    self._push_log('Task failed because of duplicate public keys.')
+                    self.push_log('Task failed because of duplicate public keys.')
                     raise AggregationError(f'find duplicate public keys: {_event.list_data}')
                 return
 
@@ -736,7 +684,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
     def _generate_shared_data(self) -> List[str]:
         """Generate information for sharing depending on user_1 list."""
-        self._push_log('Generating secret shares ...')
+        self.push_log('Generating secret shares ...')
         self._seed = secrets.token_bytes(32)
         secret_share = SecretShare(self._t)
         seed_shares = secret_share.make_shares(self._seed, len(self._users_1))
@@ -758,7 +706,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
     def _wait_for_user_2_list(self):
         """Wait for receiving the list of user_2."""
         self._users_2.clear()
-        self._push_log('Waiting for receiving users_2 list ...')
+        self.push_log('Waiting for receiving users_2 list ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, DistributeUser2ListEvent):
                 if _event.round != self._round:
@@ -767,10 +715,10 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                  for _user in _event.list_data]
                 self._users_2 = [_user for _user in self._users_2
                                  if _user.target_id == self.id]
-                self._push_log(f'Received {len(self._users_2)} users_2.')
+                self.push_log(f'Received {len(self._users_2)} users_2.')
 
                 if self._get_user_2_num() < self._t - 1:
-                    self._push_log('Task failed because of too few users_2.')
+                    self.push_log('Task failed because of too few users_2.')
                     raise AggregationError(f'too few users_2: {len(self._users_1)}')
 
                 for _user in self._users_2:
@@ -796,7 +744,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
             )
         ), f'invalid state_dict data: {state_dict}'
 
-        self._push_log('Masking parameters ...')
+        self.push_log('Masking parameters ...')
         # in order to make every calculator's PRG generates masks in a same order
         sorted_keys = sorted(state_dict)
         seed_PRG: Generator = np.random.default_rng(list(self._seed))
@@ -814,14 +762,14 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                     state_dict[_key] += _user_mask
                 else:
                     state_dict[_key] -= _user_mask
-        self._push_log('Successfully masked parameters.')
+        self.push_log('Successfully masked parameters.')
         return state_dict
 
     def _wait_for_unmasking(self):
         """Wait for unmasking updates."""
         self._users_3.clear()
 
-        self._push_log('Waiting for unmasking signal ...')
+        self.push_log('Waiting for unmasking signal ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, StartUnmaskingEvent):
                 if _event.round != self._round:
@@ -850,7 +798,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                     list_data=[_info.encode() for _info in sk_share_list],
                     aggregator=self._aggregator
                 )
-                self._push_log('Secret shares uploaded.')
+                self.push_log('Secret shares uploaded.')
 
                 return
 
